REBOL []

install-HTTPd-extension [
	name: 'mod-rsp
	verbose: 0
	
	webapps: make block! 4
		
	order: [
		url-to-filename	first
		make-response	first
	;	logging			last
		clean-up		first
		task-part		normal
		task-done		normal
		task-failed		normal
	]
	
	;====== Server events handling ======
	on-reload: does [
		set 'session-ctx reduce [
			sessions/list
			sessions/queue
		]
	]

	on-reloaded: does [
		sessions/list: session-ctx/1
		sessions/queue: session-ctx/2
		set 'session-ctx none
	]
	
	on-quit: func [svc /local flags][
		if all [
			attempt [flags: svc/conf/globals/persist]
			find flags 'sessions 
		][
			sessions/save-ctx
		]
	]
	
	on-started: does [
		sessions/load-ctx
	]
	;===================================

	send-msg: func [req sess /local job][	
		job: compose/only [
			cfg	 (req/cfg)
			in 	 (req/in)
			ip 	 (service/client/remote-ip)
			port (service/client/local-port)
			session
		]
		either sess [
			append/only job compose/only [
				id	 	(sess/id)
				vars 	(sess/vars)
				queries (sess/cache-queries)
				app	 	(req/app)
				timeout (sess/timeout)
				init 	(sess/init)
			]
			sess/init: no
			sess/busy?: yes
		][
			append job none
		]
		service/module: 'RSP		
		service/do-task mold/all job req
	]
	
	decode-msg: func [req /local ro value list sess][
		ro: req/out
		if verbose > 2 [log/info ro/content]
		
		parse first load/all ro/content [
			(ro/content: none)
			'status [
				set value string!    (
					ro/status-line: rejoin ["HTTP/1.1 " value crlf]
					ro/code: load copy/part value 3
				)
				| set value integer! (ro/code: value)
				| none!	   (ro/code: 200)
			]
			'headers [
				set list block! (
					foreach [name value] list [h-store ro/headers name value]
				)
				| none!
			]
			'content set value [string! | binary!] (ro/content: value)
			'forward set value [string! | url! | none!] (ro/forward: value)
			'session [
				into [
					opt ['init (sess: sessions/create req)]
					'id [set value string! | none!] (
						sess: any [sess sessions/obtain value]
						sessions/set-cookie sess req
					) 
					'vars set list block! (sess/vars: list)	
					'queries set list [block! | none!] (sess/cache-queries: list)
					'timeout set value time! (
						if sess/timeout <> value [
							sess/timeout: value
							sessions/refresh sess req
						]
					)
					opt ['close (sessions/destroy sess req)]
					(sess/busy?: no)
				]
				| none!
			]
		]
	]
	
	process-next-job: has [sess job req current][
		if set [sess job] sessions/pop-job [		
			req: job/1
			current: service/client
			service/set-peer job/2
			send-msg req sess
			service/set-peer current
		]
	]
	
	process-webapp: func [req /local pos url sess][
		if any [
			not sess: sessions/exists? req 
			sess/app <> req/app
		][
			sess: sessions/create req
			repend sess/vars ['login? no]
		]
		either url: select req/cfg 'auth [
			either any [
				all [pos: find/tail req/in/path "public/" tail? pos] ;find/match
				select sess/vars 'login?
			][
				; TBD: set no-cache headers to avoid browser caching of protected ressources
				if declined? req [throw false]	; let other modules serve the allowed ressource	
			][
				if not all [	; test if the requested url is not the login page
					pos: find/tail rejoin [req/app req/in/path req/in/target] url
					tail? pos
				][
					req/out/code: 302
					sessions/set-cookie sess req
					h-store req/out/headers 'Location url
					h-store req/out/headers 'Last-Modified none
					throw true 	; redirects to login page
				]
			]
		][
			if declined? req [throw false]	; no auth and not RSP ressource
		]
		sess
	]
	
	declined?: func [req]['RSP <> req/handler]
	
;=== HTTP Callbacks ===

	url-to-filename: func [req /local pos pos2][
		foreach [path cfg] webapps [
			if pos: find/match req/in/path path [			
				if any [
					req/in/target = "app-init.r"
					all [pos2: find/tail req/in/path "private/" tail? pos2]
				][
					req/out/code: 404
					return false
				]
				remove/part req/in/path pos ; remove app-name from URI
				if empty? req/in/path [req/in/path: copy "/"]
				req/app: path
				req/cfg: cfg
				break
			]
		]
		false
	]
	
	make-response: func [req /local res sess][
		if all [declined? req none? req/app][return none]	
		
		if logic? res: catch [
			if req/app [sess: process-webapp req]
			
			req/in/file: join req/in/path req/in/target 
			all [
				any [sess sess: sessions/exists? req]
				sessions/refresh sess req
				sess/busy?
				sessions/push-job sess req service/client
				return true
			]
			none
		][
			return res
		]
		
		send-msg req sess
		true
	]
	
	logging: func [req][
		none
	]

	clean-up: func [req][
		if all [declined? req none? req/app][return none]
		if zero? remainder length? sessions/list 20 [sessions/gc]
		true
	]
	
	task-part: func [req /local ro res][
		if declined? req [return none]
		ro: req/out
		ro/code: 200
		res: load/all ro/content		
		if res/headers [		
			foreach [name value] res/headers [
				h-store ro/headers name value
			]
		]
		ro/content: res/content
		service/send-chunk req
		true
	]

	task-done: func [req][
		if declined? req [return none]	;-- Check if this doesn't block some valid calls
		decode-msg req
		process-next-job
		service/process-queue
		true 		
	]

	task-failed: func [req /local ro][
		if declined? req [return none]		
		ro: req/out
		if verbose > 0 [log/info mold ro/content]
		if not any-string? ro/content [ro/content: mold ro/content]
		ro/code: 500	
		process-next-job		
		service/process-queue	
		true
	]
	
;=== Config Options ===
	
	words: [
		;--- Allow databases declaration for background tasks
		databases: [block!] in globals
		
		;--- Define a new webapp structure
		webapp: [block!] in main do [	
			repend webapps [select args/1 'virtual-root args/1]
			if not find args/1 'default [
				repend args/1 ['default [%index.rsp %index.html]]
			]
			process args/1
		]

		;--- Define a unique URI for the webapp
		virtual-root: [string!] in main do [
			if slash = last args/1 [remove back tail args/1]
		]

		;--- Trigger the authentication mode
		auth: [string!] in main		

		;--- Trigger the debugging mode
		debug: in main
		
		;--- Define the localization resources directory (have to be unique or absolute)
		locales-dir: [file!] in main
		
		;--- Optionally define a default lang (if <> 'en). Use standard language codification xx[-yy].
		default-lang: [word!] in main
		
		;--- Set session timeout 
		timeout: [time!] in main		

	]
	
;=== Session Management ===
	
	sessions: make log-class [
		name: 'RSP-Session
		verbose: 0
		
		list: make hash! 500
		queue: make block! 100 ; [sess [req client-port]...]
		
		random/seed now/time/precise

		SID-name: "RSPSID"
		ID-chars: charset [#"A" - #"Z"]
		ws: #" "
		
		ctx-file: join system/options/path %.rsp-sessions
		
		proto: context [
			id: vars: start: expires: timeout: queue: init: app: auth: 
			cache-queries: busy?: cookie: none
		]
		
		generate-ID: has [out][
			out: make string! 24
			loop 24 [insert tail out #"@" + random 26]
			uppercase out
		]
		
		save-ctx: does [
			foreach [id sess] list [
				sess/init: yes
				sess/busy?: no
				clear sess/queue
			]
			attempt [write ctx-file mold/all list]
		]
		
		load-ctx: does [	
			attempt [
				if system/words/exists? ctx-file [
					list: load ctx-file
					delete ctx-file
				]
			]
		]
		
		create: func [req /local sess][
			sess: make proto []
			sess/init: yes
			sess/id: generate-ID
			sess/vars: make block! 3
			sess/start: now
			sess/timeout: any [select req/cfg 'timeout 00:20]
			sess/expires: sess/start + sess/timeout
			sess/queue: make block! 3 ;-- [[req client-port]...]
			sess/app: req/app
			sess/auth: sess/busy?: no
			sess/cookie: build-cookie sess
			
			insert tail list sess/id
			insert tail list sess
			sess
		]
		
		exists?: func [req /local pos id sess][
			if not all [
				pos: select req/in/headers 'Cookie				
				parse/all pos [
					thru SID-name opt ws "=" opt ws
					copy id 24 ID-chars to end
				]
				sess: select list trim id 
			][
				return none
			]
			if sess/expires < now [				
				destroy sess req
				return none
			]
			sess
		]
		
		obtain: func [id][
			select list id
		]
		
		gc: does [
			remove-each [id sess] list [sess/expires < now]
		]
		
		push-job: func [sess req port /local job][
			repend/only sess/queue job: reduce [req port]
			repend queue [sess job]
		]
		
		pop-job: has [job][
			if empty? queue [return none]
			job: copy/part queue 2
			remove/part queue 2
			remove find/only job/1/queue job/2	;-- job/1=>sess, job/2=>[req port]		
			job
		]
				
		refresh: func [sess req][
			sess/expires: now + sess/timeout
		]
		
		destroy: func [sess req][
			remove/part back find list sess 2
			sess/expires: none
			if not empty? sess/queue [
				log/warn join 
					"session closed, but queue not empty : " 
					length? sess/queue
			]
		]
		
		set-cookie: func [sess req][
			h-store req/out/headers 'Set-Cookie sess/cookie
		]
		
		build-cookie: func [sess /local out][
			out: make string! 128
			insert tail out SID-name
			insert tail out "="
			insert tail out sess/id
			if sess/app [
				insert tail out "; path="
				insert tail out sess/app
			]
			insert tail out "; HttpOnly"
			out
		]
	]
]